# Version 1.0.0
bl_info = {
    "name": "Anything Can Be A Clock by Colin T (ManiacalV)",
    "blender": (2, 80, 0),
    "category": "Object",
}
import bpy
import math
from datetime import datetime, date, timedelta
import time

class ClockSettings(bpy.types.PropertyGroup):
    # XYZ method for rotation?
    axis_h : bpy.props.EnumProperty(
                name = "Hour Axis", 
                items = [
                    ("x", "X", "Rotate on X axis"),
                    ("y", "Y", "Rotate on Y axis"),
                    ("z", "Z", "Rotate on Z axis"),
                ]
    )
    axis_m : bpy.props.EnumProperty(
                name = "Minute Axis", 
                items = [
                    ("x", "X", "Rotate on X axis"),
                    ("y", "Y", "Rotate on Y axis"),
                    ("z", "Z", "Rotate on Z axis"),
                ]
    )
    axis_s : bpy.props.EnumProperty(
                name = "Second Axis", 
                items = [
                    ("x", "X", "Rotate on X axis"),
                    ("y", "Y", "Rotate on Y axis"),
                    ("z", "Z", "Rotate on Z axis"),
                ]
    )
    # is method inverted?
    axis_inv_h : bpy.props.BoolProperty(
                name = "-",
                description = "Invert direction Hour"
    )
    axis_inv_m : bpy.props.BoolProperty(
                name = "-",
                description = "Invert direction Minute"
    )
    axis_inv_s : bpy.props.BoolProperty(
                name = "-",
                description = "Invert direction Second"
    )
    
    # set up properties
    auto_clock : bpy.props.BoolProperty(default=False)
    time_scale : bpy.props.IntProperty( default=1, min=0, max=1000, name="")
    
    hour_hand_selected : bpy.props.BoolProperty(default=False)    
    hour_hand_object : bpy.props.StringProperty( default="" )
    minute_hand_selected : bpy.props.BoolProperty(default=False)
    minute_hand_object : bpy.props.StringProperty( default="" )
    second_hand_selected : bpy.props.BoolProperty(default=False)
    second_hand_object : bpy.props.StringProperty( default="" )
    
    hours : bpy.props.IntProperty( default=0)
    minutes : bpy.props.IntProperty(default=0)
    seconds : bpy.props.IntProperty(default=0)
    
    time_origin : bpy.props.StringProperty()
    time_start : bpy.props.StringProperty()
    
    user_hours : bpy.props.IntProperty(default=12, min=1, max=12, name="")
    user_minutes : bpy.props.IntProperty(default=0, min=0, max=59, name="")
    user_seconds : bpy.props.IntProperty(default=0, min=0, max=59, name="")
    
    key_start : bpy.props.IntProperty(name="", )
    key_end : bpy.props.IntProperty(name="",  )
    key_fps : bpy.props.IntProperty(name="", min=1,  )
    key_frames : bpy.props.IntProperty(name="", min=1, max=100, default=10)
    
class ADDONS_PT_ClockMakerPanel(bpy.types.Panel):
    bl_idname = "ADDONS_PT_ClockMakerPanel"
    bl_label = "Clock Maker Panel"
    bl_space_type = "VIEW_3D"   
    bl_region_type = "UI"
    bl_category = "Clock Utils"
    bl_context = "objectmode"  
    def draw(self, context):
        layout = self.layout
        obj = context.object
        settings = bpy.context.scene.clock_settings
        hour_label = "Clear Hour Hand" if settings.hour_hand_object else "Pick an Hour Hand"
        minute_label = "Clear Minute Hand" if settings.minute_hand_object else "Pick an Minute Hand"
        second_label = "Clear Second Hand" if settings.second_hand_object else "Pick an Second Hand"
        auto_icon = "CANCEL" if settings.auto_clock else "TIME"
        auto_label = "Stop Time" if settings.auto_clock else "Start Time"
        box = layout.box()
        row = box.row()
        row.label(text="Hour Hand is: " + settings.hour_hand_object)
        row = box.row()
        row.operator(OBJECT_OT_select_hour_object.bl_idname, text=str(hour_label), icon="NONE")
        row = box.row()
        row.label(text="Axis Hour", icon='NONE')
        row.prop(settings, "axis_h", expand=True)
        row.prop(settings, "axis_inv_h", toggle=True)
        
        box = layout.box()
        row = box.row()
        row.label(text="Minute Hand is: " + settings.minute_hand_object)
        row = box.row()
        row.label(text="Axis Minute", icon='NONE')
        row.prop(settings, "axis_m", expand=True)
        row.prop(settings, "axis_inv_m", toggle=True)        
        row = box.row()
        row.operator(OBJECT_OT_select_minute_object.bl_idname, text=str(minute_label), icon="NONE")
        
        box = layout.box()
        row = box.row()
        row.label(text="Second Hand is: " + settings.second_hand_object)
        row = box.row()
        row.label(text="Axis Second", icon='NONE')
        row.prop(settings, "axis_s", expand=True)
        row.prop(settings, "axis_inv_s", toggle=True)
        row = box.row()        
        row.operator(OBJECT_OT_select_second_object.bl_idname, text=str(second_label), icon="NONE")
       
        row = layout.separator() 
               
        box = layout.box()
        row = box.row()
        row.label(text="Clock Functions")
        row = box.row()
        row.operator(OBJECT_OT_update_clock.bl_idname, text=str("Set Current Time"), icon="NONE")
        row.operator(OBJECT_OT_reset_clock.bl_idname, text=str("Reset Clock"), icon="NONE")
        row.operator(OBJECT_OT_autorefresh.bl_idname, text=str(auto_label), icon=str(auto_icon) )
        box = layout.box()
        row = box.row()
        row.label(text="Set Time")
        row = box.row()
        row.prop(settings, "user_hours")
        row.prop(settings, "user_minutes")
        row.prop(settings, "user_seconds")
        row = box.row()
        row.operator(OBJECT_OT_user_set_time.bl_idname, text=str("Set Time"), icon="NONE")
        box = layout.box()
        row = box.row()
        row.label(text="Start frame")
        row.prop(settings, "key_start")
        row = box.row()
        row.label(text="End frane")
        row.prop(settings, "key_end")
        row = box.row()
        row.label(text="FPS")
        row.prop(settings, "key_fps")
        row = box.row()
        row.label(text="Keyframe every")
        row.prop(settings, "key_frames")
        row = box.row()
        row.label(text="Time Scale")
        row.prop(settings, "time_scale")
        row = box.row()
        row.operator(OBJECT_OT_create_keyframes.bl_idname, text=str("Set Keyframes"), icon="NONE")

class OBJECT_OT_user_set_time(bpy.types.Operator):
    bl_idname = "object.user_set_time"
    bl_label = "Set User Time"
    bl_description = "Set the clock time manually"
    def execute(self, context):
        settings = bpy.context.scene.clock_settings
        settings.auto_clock = False
        settings.hours      = settings.user_hours
        settings.minutes    = settings.user_minutes
        settings.seconds    = settings.user_seconds
        MoveClock(settings.hours, settings.minutes, settings.seconds)
        return {'FINISHED'}

class OBJECT_OT_update_clock(bpy.types.Operator):
    bl_idname = "object.update_clock"
    bl_label = "Update Clock"
    def execute(self, context):
        settings = bpy.context.scene.clock_settings
        settings.hours      = int(datetime.now().strftime('%I'))
        settings.minutes    = int(datetime.now().strftime('%M'))
        settings.seconds    = int(datetime.now().strftime('%S'))
        MoveClock( settings.hours, settings.minutes, settings.seconds )
        return {'FINISHED'}

class OBJECT_OT_reset_clock(bpy.types.Operator):
    bl_idname = "object.reset_clock"
    bl_label = "Reset Clock"
    bl_description = "Reset clock to start position (12:00:00)"
    def execute(self, context):
        settings = bpy.context.scene.clock_settings
        settings.auto_clock = False
        if settings.hour_hand_object:
            hourObj = bpy.context.scene.objects[settings.hour_hand_object]
            hourObj.rotation_euler = (0, 0, 0)    
            settings.hours = 0
        if settings.minute_hand_object:
            minuteObj = bpy.context.scene.objects[settings.minute_hand_object]
            minuteObj.rotation_euler = (0, 0, 0)
            settings.minutes = 0
        if settings.second_hand_object:
            secondObj = bpy.context.scene.objects[settings.second_hand_object]
            secondObj.rotation_euler = (0, 0, 0)
            settings.seconds = 0
        return {'FINISHED'}

class OBJECT_OT_autorefresh(bpy.types.Operator):
    bl_idname = "object.autorefresh"
    bl_label = "Autorefresh"
    bl_description = "Start or Stop the realtime viewport clock"
    
    def execute(self, context):
        settings = bpy.context.scene.clock_settings
        if( False == bpy.app.timers.is_registered(IncrementTime)):
            bpy.app.timers.register(IncrementTime)
            settings.auto_clock = True
            origin_string = str(settings.hours)+":"+str(settings.minutes)+":"+str(settings.seconds)
            origin = datetime.strptime(origin_string, "%H:%M:%S").strftime("%I:%M:%S")
            start = datetime.now().strftime("%I:%M:%S") 
            settings.time_origin = str(origin)
            settings.time_start = str(start)
        else:
            bpy.app.timers.unregister(IncrementTime)
            settings.auto_clock = False
        return {'FINISHED'}

class OBJECT_OT_create_keyframes(bpy.types.Operator):
    bl_idname = "object.create_keyframes"
    bl_label = "Create Keyframes"
    bl_description = "Create animation keyframes for the clock rotation based on the previous settings."
    def execute(self, context):
        settings = bpy.context.scene.clock_settings
        settings.auto_clock = False
        
        key_start = settings.key_start
        key_end = settings.key_end
        key_fps = settings.key_fps
        key_frames = settings.key_frames 
        
        if key_start >= key_end:
            ShowMessageBox("Your start keyframe must come before your end keyframe!")
            return {'FINISHED'}
        
        fps_loop = 0
        key_loop = 0
        frame_count = 0
        
        # get hand objects
        if settings.hour_hand_object:
            hourObj = bpy.context.scene.objects[settings.hour_hand_object]
        if settings.minute_hand_object:
            minuteObj = bpy.context.scene.objects[settings.minute_hand_object]
        if settings.second_hand_object:
            secondObj = bpy.context.scene.objects[settings.second_hand_object]
        
        # define angle increments
        hour_inc_raw = 6/60/60*settings.time_scale/settings.key_fps
        minute_inc_raw = 6/60*settings.time_scale/settings.key_fps
        second_inc_raw = 6*settings.time_scale/settings.key_fps
        
        # set up angle lists
        hour_angle : List(float) = [0.0,0.0,0.0]
        minute_angle : List(float) = [0.0,0.0,0.0]
        second_angle : List(float) = [0.0,0.0,0.0]
        
        # Loop each frame
        for frame_num in range( key_start, key_end+1 ):
            frame_count += 1
            
            # get current angle for each object being used
            if settings.hour_hand_object:
                hour_angle = hourObj.rotation_euler
            if settings.minute_hand_object:
                minute_angle = minuteObj.rotation_euler            
            if settings.second_hand_object:
                second_angle = secondObj.rotation_euler
            
            # Add rotation
            if settings.hour_hand_object:
                # invert increment if needed
                hour_inc = hour_inc_raw if settings.axis_inv_h else -abs(hour_inc_raw)
                
                # find axis
                match settings.axis_h:
                    case "x":
                        hour_angle.x += hour_inc
                    case "y":
                        hour_angle.y += hour_inc
                    case "z":
                        hour_angle.z += hour_inc
                        
                # rotate object
                hourObj.rotation_euler = hour_angle
                # keyframe
                if frame_count == 1 or frame_count % key_frames == 0:
                    hourObj.keyframe_insert("rotation_euler", frame=frame_num) 
                
            if settings.minute_hand_object:
                # invert increment if needed
                minute_inc = minute_inc_raw if settings.axis_inv_m else -abs(minute_inc_raw)
                
                # find axis
                match settings.axis_m:
                    case "x":
                        minute_angle.x += minute_inc
                    case "y":
                        minute_angle.y += minute_inc
                    case "z":
                        minute_angle.z += minute_inc                
                # rotate object
                minuteObj.rotation_euler = minute_angle
                # keyframe
                if frame_count == 1 or frame_count % key_frames == 0:
                    minuteObj.keyframe_insert("rotation_euler", frame=frame_num) 
                
            if settings.second_hand_object:
                # invert increment if needed
                second_inc = second_inc_raw if settings.axis_inv_h else -abs(second_inc_raw)
                
                # find axis
                match settings.axis_s:
                    case "x":
                        second_angle.x += second_inc
                    case "y":
                        second_angle.y += second_inc
                    case "z":
                        second_angle.z += second_inc
                        
                # rotate object
                secondObj.rotation_euler = second_angle
                # keyframe
                if frame_count == 1 or frame_count % key_frames == 0:
                    secondObj.keyframe_insert("rotation_euler", frame=frame_num) 
            
        # Make all keyframes CONSTANT for ticking action
        if True:
            if settings.hour_hand_object:
                fcurves = hourObj.animation_data.action.fcurves
                for fcurve in fcurves:
                    for kf in fcurve.keyframe_points:
                        kf.interpolation = 'CONSTANT'  
            if settings.minute_hand_object:
                fcurves = minuteObj.animation_data.action.fcurves
                for fcurve in fcurves:
                    for kf in fcurve.keyframe_points:
                        kf.interpolation = 'CONSTANT'  
            if settings.second_hand_object:
                fcurves = secondObj.animation_data.action.fcurves
                for fcurve in fcurves:
                    for kf in fcurve.keyframe_points:
                        kf.interpolation = 'CONSTANT'  
                
        
        return {'FINISHED'}

class OBJECT_OT_select_hour_object(bpy.types.Operator):
    bl_idname = "object.select_hour_object"
    bl_label = "Hour Hand"
    bl_description = "Selected object becomes Hour Hand, or clears the Hour Hand's object."
    def execute(self, context):
        settings = bpy.context.scene.clock_settings
        settings.auto_clock = False
        if settings.hour_hand_object:
            settings.hour_hand_selected = False
            settings.hour_hand_object = ""
        else:            
            selection_names = bpy.context.selected_objects
            if(len(selection_names) == 0):
                ShowMessageBox("Select an object to be the hour hand.")
                return {'FINISHED'}
            if(len(selection_names) > 1):
                ShowMessageBox("Select a SINGLE object to be the hour hand.")
                return {'FINISHED'}
            if selection_names[0].name == settings.minute_hand_object or selection_names[0].name == settings.second_hand_object:
                ShowMessageBox("That object is already used as a hand.")
                return {'FINISHED'}
            settings.hour_hand_object = selection_names[0].name
            settings.hour_hand_selected = True
        return {'FINISHED'}

class OBJECT_OT_select_minute_object(bpy.types.Operator):
    bl_idname = "object.select_minute_object"
    bl_label = "Minute  Hand"
    bl_description = "Selected object becomes Minute Hand, or clears the Minute Hand's object."
    def execute(self, context):
        settings = bpy.context.scene.clock_settings
        settings.auto_clock = False
        if settings.minute_hand_object:
            settings.minute_hand_selected = False
            settings.minute_hand_object = ""
        else:            
            selection_names = bpy.context.selected_objects
            if(len(selection_names) == 0):
                ShowMessageBox("Select an object to be the minute hand.")
                return {'FINISHED'}
            if(len(selection_names) > 1):
                ShowMessageBox("Select a SINGLE object to be the minute hand.")
                return {'FINISHED'}
            if selection_names[0].name == settings.hour_hand_object or selection_names[0].name == settings.second_hand_object:
                ShowMessageBox("That object is already used as a hand.")
                return {'FINISHED'}
            settings.minute_hand_object = selection_names[0].name
            settings.minute_hand_selected = True
        return {'FINISHED'}

class OBJECT_OT_select_second_object(bpy.types.Operator):
    bl_idname = "object.select_second_object"
    bl_label = "Second  Hand"
    bl_description = "Selected object becomes Second Hand, or clears the Second Hand's object."
    def execute(self, context):
        settings = bpy.context.scene.clock_settings
        settings.auto_clock = False
        if settings.second_hand_object:
            settings.second_hand_selected = False
            settings.second_hand_object = ""
        else:            
            selection_names = bpy.context.selected_objects
            if(len(selection_names) == 0):
                ShowMessageBox("Select an object to be the second hand.")
                return {'FINISHED'}
            if(len(selection_names) > 1):
                ShowMessageBox("Select a SINGLE object to be the second hand.")
                return {'FINISHED'}
            if selection_names[0].name == settings.minute_hand_object or selection_names[0].name == settings.hour_hand_object:
                ShowMessageBox("That object is already used as a hand.")
                return {'FINISHED'}
            settings.second_hand_object = selection_names[0].name
            settings.second_hand_selected = True            
        return {'FINISHED'}

def IncrementTime():
    settings = bpy.context.scene.clock_settings
    if not settings.auto_clock:
        return {'FINISHED'}
    else:
        time_origin = datetime.strptime(settings.time_origin, '%I:%M:%S')
        time_start = datetime.strptime(settings.time_start, '%I:%M:%S')
        time_now =   datetime.strptime(datetime.now().strftime('%I:%M:%S'), '%I:%M:%S')
        difference = time_now - time_start
        updated_origin = time_origin + difference
        settings.hours = updated_origin.hour
        settings.minutes = updated_origin.minute
        settings.seconds = updated_origin.second
        MoveClock(updated_origin.hour, updated_origin.minute, updated_origin.second)
        return 1.0

def MoveClock( hours, minutes, seconds ):
        settings = bpy.context.scene.clock_settings
        
        hour_x  = min_x   = sec_x   = 0
        hour_y  = min_y   = sec_y   = 0
        hour_z  = min_z   = sec_z   = 0
        
        hour_angle = minute_angle = second_angle = 0
        good_to_go = False
        # change hour
        if settings.hour_hand_object:
            hourObj = bpy.context.scene.objects[settings.hour_hand_object]
            hour_angle = (hours*30)+(minutes*0.5)
            good_to_go = True
        # change minute
        if settings.minute_hand_object:
            minuteObj = bpy.context.scene.objects[settings.minute_hand_object]
            minute_angle = (minutes*6)
            good_to_go = True
        # change second
        if settings.second_hand_object:
            secondObj = bpy.context.scene.objects[settings.second_hand_object]
            second_angle = (seconds*6)
            good_to_go = True
        
        if good_to_go == True:    
            if not settings.axis_inv_h:
                hour_angle = -abs(math.radians(hour_angle))
            else:
                hour_angle = math.radians(hour_angle)
                
            if not settings.axis_inv_m:
                minute_angle = -abs(math.radians(minute_angle))
            else:
                minute_angle = math.radians(minute_angle)
                
            if not settings.axis_inv_s:
                second_angle = -abs(math.radians(second_angle))
            else:
                second_angle = math.radians(second_angle)
            
                
                
                
        # ------------------------
        
        if settings.hour_hand_object:
            if settings.axis_h == "x":
                hourObj.rotation_euler = (hour_angle, 0, 0)
            if settings.axis_h == "y":
                hourObj.rotation_euler = (0, hour_angle, 0)
            if settings.axis_h == "z":
                hourObj.rotation_euler = (0, 0, hour_angle)                
                
        if settings.minute_hand_object:
            if settings.axis_m == "x":
                minuteObj.rotation_euler = (minute_angle, 0, 0)
            if settings.axis_m == "y":
                minuteObj.rotation_euler = (0, minute_angle, 0)
            if settings.axis_m == "z":
                minuteObj.rotation_euler = (0, 0, minute_angle)
                
                
        if settings.second_hand_object:
            if settings.axis_s == "x":
                secondObj.rotation_euler = (second_angle, 0, 0)
            if settings.axis_s == "y":
                secondObj.rotation_euler = (0, second_angle, 0)
            if settings.axis_s == "z":
                secondObj.rotation_euler = (0, 0, second_angle)
        
        if not good_to_go:
            ShowMessageBox("Select an axis to rotate on")
            return {'FINISHED'}

def ShowMessageBox(message = "", title = "Alert!", icon = 'INFO'):
    def draw(self, context):
        self.layout.label(text=message)
    bpy.context.window_manager.popup_menu(draw, title = title, icon = icon)

classes = (ADDONS_PT_ClockMakerPanel, ClockSettings, OBJECT_OT_select_hour_object, OBJECT_OT_select_minute_object, OBJECT_OT_select_second_object, OBJECT_OT_update_clock, OBJECT_OT_reset_clock, OBJECT_OT_autorefresh, OBJECT_OT_user_set_time, OBJECT_OT_create_keyframes, )

def register():    
    for cls in classes:
        bpy.utils.register_class(cls)
    bpy.types.Scene.clock_settings = bpy.props.PointerProperty(type=ClockSettings)

def unregister():
    for cls in classes:
        bpy.utils.unregister_class(cls)
    del bpy.types.Scene.clock_settings

if __name__ == "__main__":
    register()